from ete3 import Tree

def sortFunc(e):
    return e['len']

#reduce tree size for processing 
def break_up(tree,size=100):
    
    nw_trees =[]
    poss_trees = []
    
    while(len(tree.get_leaves()) > size):  
    
        poss_trees.clear()
        for n in tree.traverse():
            if len(n.get_leaves()) < size:
                poss_trees.append({'node': n, 'len': len(n.get_leaves())})
                
        poss_trees.sort(key=sortFunc)
        _n = poss_trees.pop()['node']
        print("Removing: " + str(len(_n.get_leaves())) + " | Tree_total: " + str(len(tree.get_leaves())))
        nw_trees.append(_n.detach())
        
    nw_trees.append(tree)
    return nw_trees
    

    
    



from ete3 import Tree

#@desc this function will make an attempt to attribute duplications labelled on a gene tree to
# wgd events on a species tree.
def place_wgd(species_root,gene_wgd,wgd_num=0):      
    
    children = species_root.get_children()
    
    if('WGD' in species_root.features and species_root.WGD == 'Y'):
                
        #match species tree node to gene tree
        species = []
        cmn_leaves = []
        
        for s in species_root.get_leaves():
            species.append(s.name)
            
            for g in gene_wgd.get_leaves():
                if s.name in g.name:
                    cmn_leaves.append(g)
        
        #if we were able to match leaves to the gene tree
        if cmn_leaves:
            
            nxt_wgd = gene_wgd.get_common_ancestor(cmn_leaves)
            
            leaves = []
            
            #get all the species under the chosen node on the gene tree
            for l in nxt_wgd.get_leaves():
                if l.name.split('_')[0] not in leaves:
                    leaves.append(l.name.split('_')[0])
            
            #make sure were in the right place before placing

            if sorted(species) == sorted(leaves): 
                
                c = nxt_wgd.get_children()
                
                #place event
                eventname = "event" + str(wgd_num)
                
                if 'evoltype' in nxt_wgd.features and nxt_wgd.evoltype == 'D':
                       
                    #make sure we are placing wgd events to minimize ssd events
                    if c and 'evoltype' in c[0].features and 'evoltype' in c[1].features and c[0].evoltype == 'D' and c[1].evoltype == 'D':
                        
                        place_wgd(species_root,c[0],wgd_num)
                        place_wgd(species_root,c[1],wgd_num)
                
                    else:
                        
                        nxt_wgd.add_feature(eventname,"P")
                            
                        #make recursive calls
                        if children and c:
                                                            
                            place_wgd(children[0],c[0],wgd_num+1)
                            place_wgd(children[1],c[1],wgd_num+1)
                                                            
                            place_wgd(children[1],c[0],wgd_num+1)
                            place_wgd(children[0],c[1],wgd_num+1)

                else:
                    for l in nxt_wgd.get_leaves():
                        if "_LOST" not in l.name:
                            nxt_wgd.add_feature(eventname,"M")
                            break
                    
                    if children:
                        place_wgd(children[0],gene_wgd,wgd_num+1)
                        place_wgd(children[1],gene_wgd,wgd_num+1)
                
            elif len(leaves) > len(species):
                place_wgd(species_root,gene_wgd.children[0],wgd_num)
                place_wgd(species_root,gene_wgd.children[1],wgd_num)
                
    else:
        if children:
            place_wgd(children[0],gene_wgd,wgd_num)
            place_wgd(children[1],gene_wgd,wgd_num)
    
    return 0

#count copies 
def rrates(gene_tree,pSpecies):
    
    #initialize variables
    rcopies = 0
    lcopies = 0
    l_poss = 0
    r_poss = 0
    p = 0
    
    for l in gene_tree.get_leaves():
        if pSpecies in l.name and "_LOST" not in l.name:
            p+=1

    if p!=0:
         
        for g in gene_tree.iter_descendants():
            if("event0" in g.features):
                copies = 0
                c_poss = 0
                
                for n in g.iter_descendants():
                    if "event1" in n.features:
                        c_poss += 2
                        temp = 0
                        for l in n.get_leaves():
                            if pSpecies in l.name and "_LOST" not in l.name:
                                temp+=1
                                
                        if temp > 2:
                            copies += 2
                        else:
                            copies += temp
                
                if g.event0 == 'M':
                    lcopies += copies
                    l_poss += c_poss
                
                elif g.event0 == 'P':
                    rcopies += copies
                    r_poss += c_poss

    if r_poss == 0 and l_poss == 0:
        ret = -1
    elif rcopies == 0 and lcopies == 0 and p != 0:
        ret = -1
    else:
        ret = ((rcopies,r_poss),(lcopies,l_poss),p)
    
    return ret



from ete3 import Tree
from ete3 import PhyloTree
import numpy as np



def GET_SPECIES(n):
    return n.name.split('_')[0]

def lca(G,S): # map gene tree node to species tree
    
    species = []
    s_match = []

    #get species present in gene tree 'G'
    for n in G.get_leaves():
        s = GET_SPECIES(n)
                    
        if s not in species:
            species.append(s)
            
            #match species on 'G' to a node on  'S'
            for l in S.get_leaves():
                if s in l.name:
                    s_match.append(l)

    if len(s_match) > 1:

        try:
            return S.get_common_ancestor(s_match)
        except:
            return S

    else:
            
        return s_match[0]


def find_loss(gene_tree):
    losses = 0
    
    for l in gene_tree.get_leaves():
        if "_" not in l.name:
            l.name = l.name + "_LOST"
            losses+=1
    
    return losses


def find_dups(gene_tree):

    children = []
    dups = 0
       
    for node in gene_tree.traverse():
           
        children = node.get_children()
        
        if children:
            for n in children[0].get_leaves():
                species1 = GET_SPECIES(n)
                
                for q in children[1].get_leaves():
                    species2 = GET_SPECIES(q)
                        
                    if(species1 == species2):
                        node.add_feature('evoltype','D')
                        dups += 1
                        break
            
                if 'evoltype' in node.features: 
                    break
            
            if('evoltype' not in node.features):
                node.add_feature('evoltype','S')
            
            
            
    return dups

def join(n1,n2):
    
    n0 = PhyloTree()
    n0.name = "in_node"
    
    if n1.dist == n2.dist:
        n0.dist = n1.dist/2
        n1.dist = n1.dist/2
        n2.dist = n2.dist/2
    

    else:
        n0.dist = 25
    
    n0.add_child(n1)
    n0.add_child(n2)
            
    return n0


def reconcile(G,_S):
    
    S = _S.copy()
    s_children = S.get_children()
    g_children = G.get_children()
    
    if not g_children and not s_children:
        return G
    
    #1
    if s_children:
        if(lca(G,S) in s_children[0].traverse()):
            parent = s_children[0].up
            s_children[0].detach() #replace
            parent.add_child(reconcile(G,s_children[0]))
            return S
            
        elif(lca(G,S) in s_children[1].traverse()):
            parent = s_children[1].up
            s_children[1].detach() #replace
            parent.add_child(reconcile(G,s_children[1]))
            return S
    
    #2
    if g_children and s_children:
        if(lca(G,S) == S 
           and (lca(g_children[0],S) in s_children[0].traverse()) 
           and (lca(g_children[1],S) in s_children[1].traverse())
           ):

            n1 = reconcile(g_children[0],s_children[0])
            n2 = reconcile(g_children[1],s_children[1])
            return join(n1,n2)
            
        elif(lca(G,S) == S 
             and lca(g_children[0],S) in s_children[1].traverse() 
             and lca(g_children[1],S) in s_children[0].traverse()
             ):

            n1 = reconcile(g_children[0],s_children[1])
            n2 = reconcile(g_children[1],s_children[0])
            return join(n1,n2)

    #3
    if g_children:
        if(lca(G,S)==S 
           and (lca(g_children[0],S) == S  or lca(g_children[1],S) == S)
           ):
            
            n1 = reconcile(g_children[0],S)
            n2 = reconcile(g_children[1],S)
            return join(n1,n2)
    
    print("Reconciliation failed! Maybe your tree isn't binary")
    exit()

    
def root(gene_tree,species_tree):
        
    score = 999999
    for root in gene_tree.iter_descendants():
        
        tree = gene_tree.copy()
        tree.set_outgroup(tree&root.name)
        r_tree = reconcile(tree,species_tree)
        temp = find_loss(r_tree) + find_dups(r_tree)
        if(temp < score):
            best_root = root
            score = temp

    gene_tree.set_outgroup(best_root)
    tree = reconcile(gene_tree,species_tree)
    find_loss(tree)
    find_dups(tree)
    return tree




#get all our function for easy access

from wgdtree.simulation import sim
from wgdtree.retention_rates import rrates, place_wgd
from wgdtree.rooting import root
from wgdtree.split import break_up
from wgdtree.batch_run import run



from ete3 import Tree
from wgdtree.retention_rates import rrates, place_wgd

def run(list_of_gene_trees, species_tree):

    results_dic = {}
    paired_trees = []
    pSpecies = []

    for l in species_tree.get_leaves():
        results_dic[l.name] = []

    clean_tree = species_tree.copy()

    for n in clean_tree.iter_descendants():
        if('WGD' in n.features):
            n.del_feature('WGD')

    i = 0
    for n in species_tree.traverse():

        if('WGD' in n.features and n.WGD == 'Y'):
            for x in n.iter_descendants():

                if('WGD' in x.features and x.WGD == 'Y'):

                    tmp_tree = clean_tree.copy()

                    n1 = tmp_tree&n.name

                    n2 = tmp_tree&x.name

                    n1.add_feature('WGD', 'Y')

                    n2.add_feature('WGD', 'Y')

                    paired_trees.append(tmp_tree)

                    i+=1
                    leaves = []

                    for l in x.get_leaves():
                        leaves.append(l.name.split('_')[0])

                    pSpecies.append(leaves)

    i=0
    for t in paired_trees:  #for each pair of events
        for s in pSpecies[i]: #for each species present for the pair
            l_poss = 0
            r_poss = 0
            l = 0
            r = 0
            q = 0

            for g in list_of_gene_trees: #for each gene tree
                tmp_gene = g.copy()
                place_wgd(t,tmp_gene)

                results = rrates(tmp_gene,s)

                q+=1
                if(results != -1):
                    l_poss += results[1][1]
                    r_poss += results[0][1]
                    l += results[1][0]
                    r += results[0][0]

            if(l_poss == 0 and r_poss == 0):
                results_dic[s].append((0,0))
            elif (l_poss == 0):
                results_dic[s].append((0,r/r_poss))
            elif (r_poss == 0):
                results_dic[s].append((l/l_poss,0))
            else:
                results_dic[s].append((l/l_poss,r/r_poss))



        i+=1

    return results_dic



from ete3 import Tree#, TreeStyle
from random import randint
from random import uniform
from scipy.stats import expon


global ssd_rate
global loss_rate

#=================================Helper=Functions==============================
#function for creating duplication in a tree
def dup_root(g_root,t):
    
        
    cpy = g_root.copy("newick-extended")
    cpy0 = g_root.copy("newick-extended")
    
    g_root.name = g_root.name + "_"    
        
    #fix branch dist issue
    cpy.dist = g_root.dist-t
    cpy0.dist = g_root.dist-t
    g_root.dist = t
            
    #ensure only one wgd feature 
    if("WGD" in cpy.features):
        cpy.del_feature("WGD")
        cpy0.del_feature("WGD")

    children = g_root.get_children()
    
    if children:
        #remove previous children
        children[0].detach()
        children[1].detach()
            
    #add duplicates
    g_root.add_child(cpy)
    g_root.add_child(cpy0)
        
    return

#function for creating losses in a tree
def lose_root(g_root):
        
    
    while len(g_root.up.get_children()) == 1:
        if(g_root.up):
            g_root = g_root.up
                
    g_root.detach()
            
    return

#this allows trees to be read by notung
def remove_single_links(root):
    
    children = root.get_children()
    
    if len(children) == 1:
        children[0].dist += root.dist
        parent = root.up
        parent.add_child(children[0])
        remove_single_links(children[0])
        root.detach()
        
    elif len(children) == 2:
        remove_single_links(children[0])
        remove_single_links(children[1])
    
    return
#===============================================================================

#Simulation
def bd_model(root,g_tree,c):
        
    global loss_rate
    global ssd_rate
            
    wgd_t = 99999999

    if c > 1:
        rate_sum = loss_rate + ssd_rate
    else:
        rate_sum = ssd_rate

    #create specified wgd events
    if("WGD" in root.features and root.WGD == "Y"):
        if("T" in root.features):
            wgd_t = int(root.T)
        else:
            wgd_t = int(root.dist)/2
    
    branch = root.dist

    while(branch > 0):
    
        #get time of next event
        nxt_event = expon.rvs()/(rate_sum*c)

        #do some magic to add wgd
        if wgd_t < nxt_event + (root.dist-branch) and wgd_t >= (root.dist - branch):
            
            time_of_event = wgd_t - (root.dist - branch)
            branch -= time_of_event
            
            nodes_to_dup = g_tree.search_nodes(name=root.name)
            for n in nodes_to_dup:
                c+=1
                dup_root(n,time_of_event)
                n.add_feature("WGD","Y")
                
            wgd_t = 9999999


        else:
            if branch - nxt_event > 0:
                #set rate_sum
                if c > 1:
                    rate_sum = loss_rate + ssd_rate
                else:
                    rate_sum = ssd_rate
                
                #time for some loss_rate:ssd_rate sorcery
                draw0 = uniform(0,1)
                if(draw0 < ssd_rate/rate_sum):
                    c+=1
                    #pick random copy to duplicate
                    g_poss = g_tree.search_nodes(name=root.name)
                    g_root = g_poss[randint(0,len(g_poss)-1)] 
                    
                    #duplicate
                    dup_root(g_root,nxt_event)
                else:
                    c-=1
                    #pick random copy to lose
                    g_poss = g_tree.search_nodes(name=root.name)
                    g_root = g_poss[randint(0,len(g_poss)-1)]            
                    #lose
                    lose_root(g_root)
                
            branch -= nxt_event

    children = root.get_children()

    if children:
        bd_model(children[0],g_tree,c)
        bd_model(children[1],g_tree,c)

    return

def sim(species_tree,n,path=".",ssd=.002,loss=.002):
    
    global ssd_rate
    global loss_rate
    
    
    ssd_rate = ssd
    
    loss_rate = loss
    
    
    trees = []
    for x in range(n):
        trees.append(species_tree.copy("newick-extended"))
        for q in trees[x].traverse():
            if("WGD" in  q.features):
                q.del_feature("WGD")

    num = 0

    for x in range(n):
                
        #build gene_trees and simulate losses 
        bd_model(species_tree,trees[x],1)
        remove_single_links(trees[x])

        #randomize names
        for n in trees[x].get_leaves():
            n.name = n.name + "_" + str(uniform(0,9999999))

        #write trees to files
        filename = path + "/tree" + str(num) + ".newick"
        trees[x].write(format=0,outfile=filename,features=["WGD"])
        num +=1
        
    return 0
    




